class Viewport:
    """
    A viewport is a description of surface onto which content will be
    rendered. It stores the size of the surface(in pixels), plus the
    pixel density of the viewport.
    """

    def __init__(self, width=0, height=0, dpi=None):
        self.width = width
        self.height = height
        self.dpi = dpi


class BaseBox:
    """Describe the layout of a box displaying a node.

    Stored properties
    ~~~~~~~~~~~~~~~~~
    visible: The node is included in rendering, and is visible. A value of
        False indicates the node takes up space, but is not rendered.

    content_width: The width of the content in the box
    content_height: The height of the content in the box
    content_top: The top position of the content in the box, relative to the box
    content_left: The left position of the content in the box, relative to the box
    content_bottom: The distance from the bottom of the content to the bottom of the box
    content_right: The distance from the right of the content to the right of the box

    origin_top: The absolute position of the top of the box
    origin_left: The absolute position of the left of the box

    Computed properties
    ~~~~~~~~~~~~~~~~~~~
    width: The overall width of the box
    height: The overall height of the box

    absolute_content_top: The absolute position of the top of the content box.
    absolute_content_left: The absolute position of the left of the content box.
    absolute_content_bottom: The absolute position of the bottom of the content box.
    absolute_content_right: The absolute position of the right of the content box.

    """

    def __init__(self, node):
        self.node = node
        self._reset()

    def __repr__(self):
        return "<{} ({}x{} @ {},{})>".format(
            self.__class__.__name__,
            self.content_width,
            self.content_height,
            self.absolute_content_left,
            self.absolute_content_top,
        )

    def _reset(self):
        # Some properties describing whether this node exists in
        # layout *at all*.
        self.visible = True

        # Minimum width and height of the content box.
        self.min_content_width = 0
        self.min_content_height = 0

        # Width and height of the content box.
        self.content_width = 0
        self.content_height = 0

        # Box position, relative to the containing box
        self._content_top = 0
        self._content_left = 0
        self.content_bottom = 0
        self.content_right = 0

        self.__origin_top = 0
        self.__origin_left = 0

        # Set the origin via properties; this forces the calculation of
        # absolute positions.
        self._origin_top = 0
        self._origin_left = 0

    ######################################################################
    # Origin handling
    ######################################################################
    @property
    def _origin_top(self):
        return self.__origin_top

    @_origin_top.setter
    def _origin_top(self, value):
        if value != self.__origin_top:
            self.__origin_top = value
            for child in self.node.children:
                child.layout._origin_top = self.absolute_content_top

    @property
    def _origin_left(self):
        return self.__origin_left

    @_origin_left.setter
    def _origin_left(self, value):
        if value != self.__origin_left:
            self.__origin_left = value
            for child in self.node.children:
                child.layout._origin_left = self.absolute_content_left

    @property
    def width(self):
        return self._content_left + self.content_width + self.content_right

    @property
    def min_width(self):
        return self._content_left + self.min_content_width + self.content_right

    @property
    def height(self):
        return self._content_top + self.content_height + self.content_bottom

    @property
    def min_height(self):
        return self._content_top + self.min_content_height + self.content_bottom

    ######################################################################
    # Content box properties
    ######################################################################
    @property
    def content_top(self):
        return self._content_top

    @content_top.setter
    def content_top(self, value):
        self._content_top = value
        for child in self.node.children:
            child.layout._origin_top = self.absolute_content_top

    @property
    def content_left(self):
        return self._content_left

    @content_left.setter
    def content_left(self, value):
        self._content_left = value
        for child in self.node.children:
            child.layout._origin_left = self.absolute_content_left

    ######################################################################
    # Absolute content box position
    ######################################################################

    @property
    def absolute_content_top(self):
        return self.__origin_top + self._content_top

    @property
    def absolute_content_right(self):
        return self.__origin_left + self._content_left + self.content_width

    @property
    def absolute_content_bottom(self):
        return self.__origin_top + self._content_top + self.content_height

    @property
    def absolute_content_left(self):
        return self.__origin_left + self._content_left



import pytest

from travertino.layout import BaseBox, Viewport
from travertino.node import Node
from travertino.size import BaseIntrinsicSize
from travertino.style import BaseStyle

from .utils import apply_dataclass


@apply_dataclass
class Style(BaseStyle):
    class IntrinsicSize(BaseIntrinsicSize):
        pass

    class Box(BaseBox):
        pass


def test_viewport_default():
    viewport = Viewport()

    assert viewport.width == 0
    assert viewport.height == 0
    assert viewport.dpi is None


def test_viewport_constructor():
    viewport = Viewport(width=640, height=480, dpi=96)

    assert viewport.width == 640
    assert viewport.height == 480
    assert viewport.dpi == 96


class TestBox:
    pass


@pytest.fixture
def box():
    box = TestBox()

    box.maxDiff = None

    box.grandchild1_1 = Node(style=Style())
    box.grandchild1_1.layout.min_content_width = 5
    box.grandchild1_1.layout.content_width = 10
    box.grandchild1_1.layout.min_content_height = 8
    box.grandchild1_1.layout.content_height = 16

    box.grandchild1_2 = Node(style=Style())

    box.child1 = Node(style=Style(), children=[box.grandchild1_1, box.grandchild1_2])
    box.child1.layout.min_content_width = 5
    box.child1.layout.content_width = 10
    box.child1.layout.min_content_height = 8
    box.child1.layout.content_height = 16
    box.child2 = Node(style=Style(), children=[])

    box.node = Node(style=Style(), children=[box.child1, box.child2])
    box.node.layout.min_content_width = 5
    box.node.layout.content_width = 10
    box.node.layout.min_content_height = 8
    box.node.layout.content_height = 16

    return box


def assert_layout(box, expected):
    actual = {
        "origin": (box._origin_left, box._origin_top),
        "min_size": (box.min_width, box.min_height),
        "size": (box.width, box.height),
        "content": (box.content_width, box.content_height),
        "relative": (
            box.content_top,
            box.content_right,
            box.content_bottom,
            box.content_left,
        ),
        "absolute": (
            box.absolute_content_top,
            box.absolute_content_right,
            box.absolute_content_bottom,
            box.absolute_content_left,
        ),
    }
    assert actual == expected


def test_repr(box):
    box.node.layout._origin_top = 1
    box.node.layout._origin_left = 2
    assert repr(box.node.layout) == "<Box (10x16 @ 2,1)>"


def test_initial(box):
    # Core attributes have been stored
    assert_layout(
        box.node.layout,
        {
            "origin": (0, 0),
            "min_size": (5, 8),
            "size": (10, 16),
            "content": (10, 16),
            "relative": (0, 0, 0, 0),
            "absolute": (0, 10, 16, 0),
        },
    )


@pytest.mark.parametrize(
    "dimension, val1, expected1, val2, expected2",
    [
        (
            "content_top",
            5,
            {
                "origin": (0, 0),
                "min_size": (5, 13),
                "size": (10, 21),
                "content": (10, 16),
                "relative": (5, 0, 0, 0),
                "absolute": (5, 10, 21, 0),
            },
            7,
            {
                "origin": (0, 0),
                "min_size": (5, 15),
                "size": (10, 23),
                "content": (10, 16),
                "relative": (7, 0, 0, 0),
                "absolute": (7, 10, 23, 0),
            },
        ),
        (
            "content_left",
            5,
            {
                "origin": (0, 0),
                "min_size": (10, 8),
                "size": (15, 16),
                "content": (10, 16),
                "relative": (0, 0, 0, 5),
                "absolute": (0, 15, 16, 5),
            },
            7,
            {
                "origin": (0, 0),
                "min_size": (12, 8),
                "size": (17, 16),
                "content": (10, 16),
                "relative": (0, 0, 0, 7),
                "absolute": (0, 17, 16, 7),
            },
        ),
        (
            "min_content_width",
            8,
            {
                "origin": (0, 0),
                "min_size": (8, 8),
                "size": (10, 16),
                "content": (10, 16),
                "relative": (0, 0, 0, 0),
                "absolute": (0, 10, 16, 0),
            },
            9,
            {
                "origin": (0, 0),
                "min_size": (9, 8),
                "size": (10, 16),
                "content": (10, 16),
                "relative": (0, 0, 0, 0),
                "absolute": (0, 10, 16, 0),
            },
        ),
        (
            "content_width",
            5,
            {
                "origin": (0, 0),
                "min_size": (5, 8),
                "size": (5, 16),
                "content": (5, 16),
                "relative": (0, 0, 0, 0),
                "absolute": (0, 5, 16, 0),
            },
            7,
            {
                "origin": (0, 0),
                "min_size": (5, 8),
                "size": (7, 16),
                "content": (7, 16),
                "relative": (0, 0, 0, 0),
                "absolute": (0, 7, 16, 0),
            },
        ),
        (
            "min_content_height",
            7,
            {
                "origin": (0, 0),
                "min_size": (5, 7),
                "size": (10, 16),
                "content": (10, 16),
                "relative": (0, 0, 0, 0),
                "absolute": (0, 10, 16, 0),
            },
            8,
            {
                "origin": (0, 0),
                "min_size": (5, 8),
                "size": (10, 16),
                "content": (10, 16),
                "relative": (0, 0, 0, 0),
                "absolute": (0, 10, 16, 0),
            },
        ),
        (
            "content_height",
            10,
            {
                "origin": (0, 0),
                "min_size": (5, 8),
                "size": (10, 10),
                "content": (10, 10),
                "relative": (0, 0, 0, 0),
                "absolute": (0, 10, 10, 0),
            },
            12,
            {
                "origin": (0, 0),
                "min_size": (5, 8),
                "size": (10, 12),
                "content": (10, 12),
                "relative": (0, 0, 0, 0),
                "absolute": (0, 10, 12, 0),
            },
        ),
    ],
)
def test_set_content_dimension(box, dimension, val1, expected1, val2, expected2):
    setattr(box.node.layout, dimension, val1)
    assert_layout(box.node.layout, expected1)

    # Set to a new value
    setattr(box.node.layout, dimension, val2)
    assert_layout(box.node.layout, expected2)


def test_descendent_offsets(box):
    box.node.layout.content_top = 7
    box.node.layout.content_left = 8

    box.child1.layout.content_top = 9
    box.child1.layout.content_left = 10

    box.grandchild1_1.layout.content_top = 11
    box.grandchild1_1.layout.content_left = 12

    assert_layout(
        box.node.layout,
        {
            "origin": (0, 0),
            "min_size": (13, 15),
            "size": (18, 23),
            "content": (10, 16),
            "relative": (7, 0, 0, 8),
            "absolute": (7, 18, 23, 8),
        },
    )

    assert_layout(
        box.child1.layout,
        {
            "origin": (8, 7),
            "min_size": (15, 17),
            "size": (20, 25),
            "content": (10, 16),
            "relative": (9, 0, 0, 10),
            "absolute": (16, 28, 32, 18),
        },
    )

    assert_layout(
        box.grandchild1_1.layout,
        {
            "origin": (18, 16),
            "min_size": (17, 19),
            "size": (22, 27),
            "content": (10, 16),
            "relative": (11, 0, 0, 12),
            "absolute": (27, 40, 43, 30),
        },
    )

    # Modify the grandchild position
    box.grandchild1_1.layout.content_top = 13
    box.grandchild1_1.layout.content_left = 14

    # Only the grandchild position has changed.
    assert_layout(
        box.node.layout,
        {
            "origin": (0, 0),
            "min_size": (13, 15),
            "size": (18, 23),
            "content": (10, 16),
            "relative": (7, 0, 0, 8),
            "absolute": (7, 18, 23, 8),
        },
    )

    assert_layout(
        box.child1.layout,
        {
            "origin": (8, 7),
            "min_size": (15, 17),
            "size": (20, 25),
            "content": (10, 16),
            "relative": (9, 0, 0, 10),
            "absolute": (16, 28, 32, 18),
        },
    )

    assert_layout(
        box.grandchild1_1.layout,
        {
            "origin": (18, 16),
            "min_size": (19, 21),
            "size": (24, 29),
            "content": (10, 16),
            "relative": (13, 0, 0, 14),
            "absolute": (29, 42, 45, 32),
        },
    )

    # Modify the child position
    box.child1.layout.content_top = 15
    box.child1.layout.content_left = 16

    # The child and grandchild positions have changed.
    assert_layout(
        box.node.layout,
        {
            "origin": (0, 0),
            "min_size": (13, 15),
            "size": (18, 23),
            "content": (10, 16),
            "relative": (7, 0, 0, 8),
            "absolute": (7, 18, 23, 8),
        },
    )

    assert_layout(
        box.child1.layout,
        {
            "origin": (8, 7),
            "min_size": (21, 23),
            "size": (26, 31),
            "content": (10, 16),
            "relative": (15, 0, 0, 16),
            "absolute": (22, 34, 38, 24),
        },
    )

    assert_layout(
        box.grandchild1_1.layout,
        {
            "origin": (24, 22),
            "min_size": (19, 21),
            "size": (24, 29),
            "content": (10, 16),
            "relative": (13, 0, 0, 14),
            "absolute": (35, 48, 51, 38),
        },
    )


def test_absolute_equalities(box):
    # Move the box around and set some borders.
    layout = box.node.layout

    layout.origin_top = 100
    layout.origin_left = 200

    layout.content_top = 50
    layout.content_left = 75
    layout.content_right = 42
    layout.content_bottom = 37

    assert (
        layout.absolute_content_left + layout.content_width
        == layout.absolute_content_right
    )
    assert (
        layout.absolute_content_top + layout.content_height
        == layout.absolute_content_bottom
    )

    assert (
        layout.content_left + layout.content_width + layout.content_right
        == layout.width
    )
    assert (
        layout.content_top + layout.content_height + layout.content_bottom
        == layout.height
    )



def _package_version(file, name):
    try:
        # Read version from SCM metadata
        # This will only exist in a development environment
        from setuptools_scm import get_version

        # Excluded from coverage because a pure test environment (such as the one
        # used by tox in CI) won't have setuptools_scm
        return get_version(root="../../..", relative_to=file)  # pragma: no cover
    except (
        ModuleNotFoundError,
        LookupError,
    ):  # pragma: no-cover-if-missing-setuptools_scm
        # If setuptools_scm isn't in the environment, the call to import will fail.
        # If it *is* in the environment, but the code isn't a git checkout (e.g.,
        # it's been pip installed non-editable) the call to get_version() will fail.
        # If either of these occurs, read version from the installer metadata.
        import importlib.metadata

        # The Toga package names as defined in pyproject.toml all use dashes.
        package = name.replace("_", "-")
        return importlib.metadata.version(package)


__version__ = _package_version(__file__, __name__)



from unittest.mock import Mock, call
from warnings import catch_warnings, filterwarnings

import pytest

from travertino.layout import BaseBox, Viewport
from travertino.node import Node
from travertino.properties.validated import validated_property
from travertino.size import BaseIntrinsicSize
from travertino.style import BaseStyle

from .utils import apply_dataclass, mock_apply


@mock_apply
@apply_dataclass
class Style(BaseStyle):
    int_prop: int = validated_property(integer=True)

    class IntrinsicSize(BaseIntrinsicSize):
        pass

    class Box(BaseBox):
        pass

    def layout(self, viewport):
        # A simple layout scheme that allocates twice the viewport size.
        self._applicator.node.layout.content_width = viewport.width * 2
        self._applicator.node.layout.content_height = viewport.height * 2


@mock_apply
@apply_dataclass
class OldStyle(Style):
    # Uses two-argument layout(), as in Toga <= 0.4.8
    def layout(self, node, viewport):
        # A simple layout scheme that allocates twice the viewport size.
        super().layout(viewport)


@mock_apply
@apply_dataclass
class TypeErrorStyle(Style):
    # Uses the correct signature, but raises an unrelated TypeError in layout
    def layout(self, viewport):
        raise TypeError("An unrelated TypeError has occurred somewhere in layout()")


@mock_apply
@apply_dataclass
class OldTypeErrorStyle(Style):
    # Just to be extra safe...
    def layout(self, node, viewport):
        raise TypeError("An unrelated TypeError has occurred somewhere in layout()")


@apply_dataclass
class BrokenStyle(BaseStyle):
    def apply(self):
        raise AttributeError("Missing attribute, node not ready for style application")

    class IntrinsicSize(BaseIntrinsicSize):
        pass

    class Box(BaseBox):
        pass

    def layout(self, viewport):
        # A simple layout scheme that allocates twice the viewport size.
        self._applicator.node.layout.content_width = viewport.width * 2
        self._applicator.node.layout.content_height = viewport.height * 2


@apply_dataclass
class AttributeTestStyle(BaseStyle):
    class IntrinsicSize(BaseIntrinsicSize):
        pass

    class Box(BaseBox):
        pass

    def apply(self):
        assert self._applicator.node.style is self


def test_create_leaf():
    """A leaf can be created"""
    style = Style()
    leaf = Node(style=style)

    assert leaf._children is None
    assert leaf.children == []
    assert not leaf.can_have_children

    # An unattached leaf is a root
    assert leaf.parent is None
    assert leaf.root == leaf

    # A leaf can't have children
    child = Node(style=style)

    with pytest.raises(ValueError):
        leaf.add(child)


def test_create_node():
    """A node can be created with children"""
    style = Style()

    child1 = Node(style=style)
    child2 = Node(style=style)
    child3 = Node(style=style)

    node = Node(style=style, children=[child1, child2, child3])

    assert node.children == [child1, child2, child3]
    assert node.can_have_children

    # The node is the root as well.
    assert node.parent is None
    assert node.root == node

    # The children all point at the node.
    assert child1.parent == node
    assert child1.root == node

    assert child2.parent == node
    assert child2.root == node

    assert child3.parent == node
    assert child3.root == node

    # Create another node
    new_node = Node(style=style, children=[])

    assert new_node.children == []
    assert new_node.can_have_children

    # Add the old node as a child of the new one.
    new_node.add(node)

    # The new node is the root
    assert new_node.parent is None
    assert new_node.root == new_node

    # The node is the root as well.
    assert node.parent == new_node
    assert node.root == new_node

    # The children all point at the node.
    assert child1.parent == node
    assert child1.root == new_node

    assert child2.parent == node
    assert child2.root == new_node

    assert child3.parent == node
    assert child3.root == new_node


@pytest.mark.parametrize("StyleClass", [Style, OldStyle])
def test_refresh(StyleClass):
    """The layout can be refreshed, and the applicator invoked"""

    # Define an applicator that tracks the node being rendered and its size
    class Applicator:
        def __init__(self, node):
            self.tasks = []
            self.node = node

        def set_bounds(self):
            self.tasks.append(
                (
                    self.node,
                    self.node.layout.content_width,
                    self.node.layout.content_height,
                )
            )

    class TestNode(Node):
        def __init__(self, style, children=None):
            super().__init__(
                style=style, applicator=Applicator(self), children=children
            )

    # Define a simple 2 level tree of nodes.
    style = StyleClass()
    child1 = TestNode(style=style)
    child2 = TestNode(style=style)
    child3 = TestNode(style=style)

    node = TestNode(style=style, children=[child1, child2, child3])

    # Refresh the root node
    node.refresh(Viewport(width=10, height=20))

    # Check the output is as expected
    assert node.applicator.tasks == [(node, 20, 40)]
    assert child1.applicator.tasks == []
    assert child2.applicator.tasks == []
    assert child3.applicator.tasks == []

    # Reset the applicator
    node.applicator.tasks = []

    # Refresh a child node
    child1.refresh(Viewport(width=15, height=25))

    # The root node was rendered, not the child.
    assert node.applicator.tasks == [(node, 30, 50)]
    assert child1.applicator.tasks == []
    assert child2.applicator.tasks == []
    assert child3.applicator.tasks == []


def test_refresh_no_op():
    """Refresh is a no-op if no applicator is set."""
    node = Node(style=Style())
    node.refresh(Viewport(width=100, height=100))
    node.style.apply.assert_not_called()


@pytest.mark.parametrize("StyleClass", [TypeErrorStyle, OldTypeErrorStyle])
def test_type_error_in_layout(StyleClass):
    """The shim shouldn't hide unrelated TypeErrors."""

    class Applicator:
        def set_bounds(self):
            pass

    node = Node(style=StyleClass(), applicator=Applicator())
    with pytest.raises(TypeError, match=r"unrelated TypeError"):
        node.refresh(Viewport(50, 50))


def test_add():
    """Nodes can be added as children to another node"""

    style = Style()
    node = Node(style=style, children=[])

    child = Node(style=style)
    node.add(child)

    assert child in node.children
    assert child.parent == node
    assert child.root == node.root


def test_insert():
    """Node can be inserted at a specific position as a child"""

    style = Style()
    child1 = Node(style=style)
    child2 = Node(style=style)
    child3 = Node(style=style)
    node = Node(style=style, children=[child1, child2, child3])

    child4 = Node(style=style)

    index = 2
    node.insert(index, child4)

    assert child4 in node.children
    assert child4.parent == node
    assert child4.root == node.root

    assert node.children.index(child4) == index


def test_insert_leaf():
    """Node that can't contain children raises error on insert."""
    node = Node(style=Style())
    child = Node(style=Style())
    with pytest.raises(ValueError, match=r"Cannot insert child"):
        node.insert(0, child)


def test_remove():
    """Children can be removed from node"""

    style = Style()
    child1 = Node(style=style)
    child2 = Node(style=style)
    child3 = Node(style=style)
    node = Node(style=style, children=[child1, child2, child3])

    node.remove(child1)

    assert child1 not in node.children
    assert child1.parent is None
    assert child1.root == child1


def test_remove_leaf():
    """Node that can't contain children raises error on remove."""
    node = Node(style=Style())
    child = Node(style=Style())
    with pytest.raises(ValueError, match=r"Cannot remove children"):
        node.remove(child)


def test_clear():
    """Node can be cleared of all children."""
    style = Style()
    children = [Node(style=style), Node(style=style), Node(style=style)]
    node = Node(style=style, children=children)

    for child in children:
        assert child in node.children
        assert child.parent == node
        assert child.root == node
    assert node.children == children

    node.clear()

    for child in children:
        assert child not in node.children
        assert child.parent is None
        assert child.root == child

    assert node.children == []


def test_clear_leaf():
    """For a node that can't have children, clear() is a no-op."""
    node = Node(style=Style())
    assert node.children == []
    node.clear()
    assert node.children == []


def test_create_with_no_applicator():
    """A node can be created without an applicator."""
    style = Style(int_prop=5)
    node = Node(style=style)

    # Style copies on assignment.
    assert isinstance(node.style, Style)
    assert node.style == style
    assert node.style.int_prop == 5
    assert node.style is not style

    # Since no applicator has been assigned, the overall style wasn't applied. However,
    # apply("int_prop") was still called.
    node.style.apply.assert_called_once_with("int_prop")


def test_create_with_applicator():
    """A node can be created with an applicator."""
    style = Style(int_prop=5)
    applicator = Mock()
    node = Node(style=style, applicator=applicator)

    # Style copies on assignment.
    assert isinstance(node.style, Style)
    assert node.style == style
    assert node.style.int_prop == 5
    assert node.style is not style

    # Applicator assignment does *not* copy.
    assert node.applicator is applicator
    # Applicator gets a reference back to its node and to the style.
    assert applicator.node is node
    assert node.style._applicator is applicator

    # First, call("int_prop") is called when style object is created.
    # Assigning a non-None applicator should always apply style.
    assert node.style.apply.mock_calls == [call("int_prop"), call()]


@pytest.mark.parametrize(
    "node",
    [
        Node(style=Style()),
        Node(style=Style(), applicator=Mock()),
    ],
)
def test_assign_applicator(node):
    """A node can be assigned an applicator after creation."""
    node.style.apply.reset_mock()

    applicator = Mock()
    node.applicator = applicator

    # Applicator assignment does *not* copy.
    assert node.applicator is applicator
    # Applicator gets a reference back to its node and to the style.
    assert applicator.node is node
    assert node.style._applicator is applicator

    # Assigning a non-None applicator should always apply style.
    node.style.apply.assert_called_once_with()


@pytest.mark.parametrize(
    "node",
    [
        Node(style=Style()),
        Node(style=Style(), applicator=Mock()),
    ],
)
def test_assign_applicator_none(node):
    """A node can have its applicator set to None."""
    node.style.apply.reset_mock()

    node.applicator = None
    assert node.applicator is None

    # Should be updated on style as well
    assert node.style._applicator is None
    # Assigning None to applicator does not trigger apply.
    node.style.apply.assert_not_called()


def assign_new_applicator():
    """Assigning a new applicator clears reference to node on the old applicator."""
    applicator_1 = Mock()
    node = Node(style=Style(), applicator=applicator_1)

    assert applicator_1.node is node

    applicator_2 = Mock()
    node.applicator = applicator_2

    assert applicator_1.node is None
    assert applicator_2.node is node


def assign_new_applicator_none():
    """Assigning None to applicator clears reference to node on the old applicator."""
    applicator = Mock()
    node = Node(style=Style(), applicator=applicator)

    assert applicator.node is node

    node.applicator = None

    assert applicator.node is None


def test_assign_style_with_applicator():
    """Assigning a new style triggers an apply if an applicator is already present."""
    style_1 = Style(int_prop=5)
    node = Node(style=style_1, applicator=Mock())

    style_2 = Style(int_prop=10)
    node.style = style_2

    # Style copies on assignment.
    assert isinstance(node.style, Style)
    assert node.style == style_2
    assert node.style.int_prop == 10
    assert node.style is not style_2

    assert node.style != style_1

    # call("int_prop") is called when the style object is created.
    # Since an applicator has already been assigned, assigning style applies the style.
    assert node.style.apply.mock_calls == [call("int_prop"), call()]


def test_assign_style_with_no_applicator():
    """Assigning new style doesn't trigger an apply if an applicator isn' present."""
    style_1 = Style(int_prop=5)
    node = Node(style=style_1)

    style_2 = Style(int_prop=10)
    node.style = style_2

    # Style copies on assignment.
    assert isinstance(node.style, Style)
    assert node.style == style_2
    assert node.style.int_prop == 10
    assert node.style is not style_2

    assert node.style != style_1

    # Since no applicator has been assigned, the overall style wasn't applied. However,
    # apply("int_prop") was still called when creating the style.
    node.style.apply.assert_called_once_with("int_prop")


def test_apply_before_node_is_ready():
    """Triggering an apply raises a warning if the node is not ready to apply style."""
    style = BrokenStyle()
    applicator = Mock()

    with pytest.warns(RuntimeWarning):
        node = Node(style=style)
        node.applicator = applicator

    with pytest.warns(RuntimeWarning):
        node.style = BrokenStyle()

    with pytest.warns(RuntimeWarning):
        Node(style=style, applicator=applicator)


def test_applicator_has_node_reference():
    """Applicator should have a reference to its node before style is first applied."""

    # We can't just check it after creating the widget, because at that point the
    # apply will have already happened. AttributeTestStyle has an apply() method
    # that asserts the reference trail of style -> applicator -> node -> style is
    # already intact at the point that apply is called.

    with catch_warnings():
        filterwarnings("error", category=RuntimeWarning)
        Node(style=AttributeTestStyle(), applicator=Mock())



from typing import NamedTuple

import pytest

from travertino.size import BaseIntrinsicSize, at_least


class Size(NamedTuple):
    width: int
    height: int


BASE_SIZE = Size(width=1, height=2)


class TestBox:
    pass


@pytest.fixture
def box():
    box = TestBox()

    box.maxDiff = None

    box.size = BaseIntrinsicSize()
    box.size.width, box.size.height = BASE_SIZE

    assert_size(box.size, BASE_SIZE)

    return box


def assert_size(size, values):
    assert (size.width, size.height) == values


def test_at_least_repr():
    assert repr(at_least(10)) == "at least 10"


def test_at_least_eq():
    assert at_least(10) == at_least(10)
    assert at_least(10) != at_least(4)
    assert at_least(10) != "something else"


def test_size_repr(box):
    assert repr(box.size) == "(1, 2)"
    box.size.width = at_least(10)
    assert repr(box.size) == "(at least 10, 2)"



class Node:
    def __init__(self, style, applicator=None, children=None):
        # Parent needs to be primed before style is (potentially) applied with
        # assignment of applicator.
        self._parent = None
        self._root = None

        # Explicitly set the internal attribute first, since the setter for style will
        # access the applicator property.
        self._applicator = None

        self.style = style
        self.applicator = applicator

        if children is None:
            self._children = None
        else:
            self._children = []
            for child in children:
                self.add(child)

    @property
    def style(self):
        """The node's style.

        Assigning a style triggers an application of that style if an applicator has
        already been assigned.
        """
        return self._style

    @style.setter
    def style(self, style):
        self._style = style.copy()
        self.intrinsic = self.style.IntrinsicSize()
        self.layout = self.style.Box(self)

        if self.applicator:
            self.style._applicator = self.applicator

    @property
    def applicator(self):
        """This node's applicator, which handles applying the style.

        Assigning an applicator triggers an application of the node's style.
        """
        return self._applicator

    @applicator.setter
    def applicator(self, applicator):
        if self.applicator:
            # If an existing applicator is present, clear its reference to this node.
            self.applicator.node = None

        if applicator:
            # This needs to happen *before* assigning the applicator to the style,
            # below, because as part of receiving the applicator, the style will apply
            # itself. How this happens will vary with applicator implementation, but
            # will probably need access to the node.
            applicator.node = self

        self._applicator = applicator
        # This triggers style.apply():
        self.style._applicator = applicator

    @property
    def root(self):
        """The root of the tree containing this node.

        Returns:
            The root node. Returns self if this node *is* the root node.
        """
        return self._root if self._root else self

    @property
    def parent(self):
        """The parent of this node.

        Returns:
            The parent of this node. Returns None if this node is the root node.
        """
        return self._parent

    @property
    def children(self):
        """The children of this node.
        This *always* returns a list, even if the node is a leaf
        and cannot have children.

        Returns:
            A list of the children for this widget.
        """
        if self._children is None:
            return []
        else:
            return self._children

    @property
    def can_have_children(self):
        """Determine if the node can have children.

        This does not resolve whether there actually *are* any children;
        it only confirms whether children are theoretically allowed.
        """
        return self._children is not None

    def add(self, child):
        """Add a node as a child of this one.
        Args:
            child: A node to add as a child to this node.

        Raises:
            ValueError: If this node is a leaf, and cannot have children.
        """
        if self._children is None:
            raise ValueError("Cannot add children")

        self._children.append(child)
        child._parent = self
        self._set_root(child, self.root)

    def insert(self, index, child):
        """Insert a node as a child of this one.
        Args:
            index: Index of child position.
            child: A node to insert as a child to this node.

        Raises:
            ValueError: If this node is a leaf, and cannot have children.
        """
        if self._children is None:
            raise ValueError("Cannot insert child")

        self._children.insert(index, child)
        child._parent = self
        self._set_root(child, self.root)

    def remove(self, child):
        """Remove child from this node.
        Args:
            child: The child to remove from this node.

        Raises:
            ValueError: If this node is a leaf, and cannot have children.
        """
        if self._children is None:
            raise ValueError("Cannot remove children")

        self._children.remove(child)
        child._parent = None
        self._set_root(child, None)

    def clear(self):
        """Clear all children from this node."""
        if self._children is None:
            # This is a leaf, so do nothing.
            return

        for child in self._children:
            child._parent = None
            self._set_root(child, None)
        self._children = []

    def refresh(self, viewport):
        """Refresh the layout and appearance of the tree this node is contained in."""
        if self._root:
            self._root.refresh(viewport)
        else:
            if self.applicator:

                ######################################################################
                # 2024-12: Backwards compatibility for Toga <= 0.4.8
                ######################################################################
                # Accommodate the earlier signature of layout(), which included the node
                # as a parameter.
                try:
                    self.style.layout(viewport)
                except TypeError as error:
                    if "layout() missing 1 required positional argument:" in str(error):
                        self.style.layout(self, viewport)
                    else:
                        raise
                ######################################################################
                # End backwards compatibility
                ######################################################################

                self.applicator.set_bounds()

    def _set_root(self, node, root):
        # Propagate a root node change through a tree.
        node._root = root
        for child in node.children:
            self._set_root(child, root)



from __future__ import annotations

from warnings import catch_warnings, filterwarnings

import pytest

from travertino.colors import NAMED_COLOR, rgb
from travertino.constants import GOLDENROD, NONE, REBECCAPURPLE, TOP
from travertino.properties.choices import Choices
from travertino.properties.validated import validated_property
from travertino.style import BaseStyle

from .utils import apply_dataclass, mock_apply


@mock_apply
@apply_dataclass
class Style(BaseStyle):
    none: str = validated_property(NONE, REBECCAPURPLE, initial=NONE)
    allow_string: str = validated_property(string=True, initial="start")
    allow_integer: int = validated_property(integer=True, initial=0)
    allow_number: float = validated_property(number=True, initial=0)
    allow_color: str = validated_property(color=True, initial="goldenrod")
    values: str = validated_property("a", "b", NONE, initial="a")
    multiple_choices: str | float = validated_property(
        "a",
        "b",
        NONE,
        number=True,
        color=True,
        initial=None,
    )
    string_symbol: str = validated_property(TOP, NONE)


with catch_warnings():
    filterwarnings("ignore", category=DeprecationWarning)

    @mock_apply
    class DeprecatedStyle(BaseStyle):
        pass

    DeprecatedStyle.validated_property(
        "none", choices=Choices(NONE, REBECCAPURPLE), initial=NONE
    )
    DeprecatedStyle.validated_property(
        "allow_string", choices=Choices(string=True), initial="start"
    )
    DeprecatedStyle.validated_property(
        "allow_integer", choices=Choices(integer=True), initial=0
    )
    DeprecatedStyle.validated_property(
        "allow_number", choices=Choices(number=True), initial=0
    )
    DeprecatedStyle.validated_property(
        "allow_color", choices=Choices(color=True), initial="goldenrod"
    )
    DeprecatedStyle.validated_property(
        "values", choices=Choices("a", "b", NONE), initial="a"
    )
    DeprecatedStyle.validated_property(
        "multiple_choices",
        choices=Choices("a", "b", NONE, number=True, color=True),
        initial=None,
    )
    DeprecatedStyle.validated_property("string_symbol", choices=Choices(TOP, NONE))


def assert_property(obj, name, value):
    assert getattr(obj, name) == value

    obj.apply.assert_called_once_with(name)
    obj.apply.reset_mock()


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_none(StyleClass):
    style = StyleClass()
    assert style.none == NONE

    with pytest.raises(ValueError):
        style.none = 10

    with pytest.raises(ValueError):
        style.none = 3.14159

    with pytest.raises(ValueError):
        style.none = "#112233"

    with pytest.raises(ValueError):
        style.none = "a"

    with pytest.raises(ValueError):
        style.none = "b"

    # Set the property to a different explicit value
    style.none = REBECCAPURPLE
    assert_property(style, "none", REBECCAPURPLE)

    # A Travertino NONE is an explicit value
    style.none = NONE
    assert_property(style, "none", NONE)

    # Set the property to a different explicit value
    style.none = REBECCAPURPLE
    assert_property(style, "none", REBECCAPURPLE)

    # A Python None is invalid
    with pytest.raises(ValueError):
        style.none = None

    # The property can be reset
    del style.none
    assert_property(style, "none", NONE)

    with pytest.raises(
        ValueError,
        match=r"Invalid value 'invalid' for property none; Valid values are: "
        r"none, rebeccapurple",
    ):
        style.none = "invalid"


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_allow_string(StyleClass):
    style = StyleClass()
    assert style.allow_string == "start"

    with pytest.raises(ValueError):
        style.allow_string = 10

    with pytest.raises(ValueError):
        style.allow_string = 3.14159

    style.allow_string = REBECCAPURPLE
    assert_property(style, "allow_string", "rebeccapurple")

    style.allow_string = "#112233"
    assert_property(style, "allow_string", "#112233")

    style.allow_string = "a"
    assert_property(style, "allow_string", "a")

    style.allow_string = "b"
    assert_property(style, "allow_string", "b")

    # A Travertino NONE is an explicit string value
    style.allow_string = NONE
    assert_property(style, "allow_string", NONE)

    # A Python None is invalid
    with pytest.raises(ValueError):
        style.allow_string = None

    # The property can be reset
    del style.allow_string
    assert_property(style, "allow_string", "start")

    with pytest.raises(
        ValueError,
        match=r"Invalid value 99 for property allow_string; Valid values are: <string>",
    ):
        style.allow_string = 99


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_allow_integer(StyleClass):
    style = StyleClass()
    assert style.allow_integer == 0

    style.allow_integer = 10
    assert_property(style, "allow_integer", 10)

    # This is an odd case; Python happily rounds floats to integers.
    # It's more trouble than it's worth to correct this.
    style.allow_integer = 3.14159
    assert_property(style, "allow_integer", 3)

    with pytest.raises(ValueError):
        style.allow_integer = REBECCAPURPLE

    with pytest.raises(ValueError):
        style.allow_integer = "#112233"

    with pytest.raises(ValueError):
        style.allow_integer = "a"

    with pytest.raises(ValueError):
        style.allow_integer = "b"

    # A Travertino NONE is an explicit string value
    with pytest.raises(ValueError):
        style.allow_integer = NONE

    # A Python None is invalid
    with pytest.raises(ValueError):
        style.allow_integer = None

    # The property can be reset
    del style.allow_integer
    assert_property(style, "allow_integer", 0)

    # Check the error message
    with pytest.raises(
        ValueError,
        match=(
            r"Invalid value 'invalid' for property allow_integer; Valid values are: "
            r"<integer>"
        ),
    ):
        style.allow_integer = "invalid"


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_allow_number(StyleClass):
    style = StyleClass()
    assert style.allow_number == 0

    style.allow_number = 10
    assert_property(style, "allow_number", 10.0)

    style.allow_number = 3.14159
    assert_property(style, "allow_number", 3.14159)

    with pytest.raises(ValueError):
        style.allow_number = REBECCAPURPLE

    with pytest.raises(ValueError):
        style.allow_number = "#112233"

    with pytest.raises(ValueError):
        style.allow_number = "a"

    with pytest.raises(ValueError):
        style.allow_number = "b"

    # A Travertino NONE is an explicit string value
    with pytest.raises(ValueError):
        style.allow_number = NONE

    # A Python None is invalid
    with pytest.raises(ValueError):
        style.allow_number = None

    # The property can be reset
    del style.allow_number
    assert_property(style, "allow_number", 0)

    with pytest.raises(
        ValueError,
        match=(
            r"Invalid value 'invalid' for property allow_number; Valid values are: "
            r"<number>"
        ),
    ):
        style.allow_number = "invalid"


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_allow_color(StyleClass):
    style = StyleClass()
    assert style.allow_color == NAMED_COLOR[GOLDENROD]

    with pytest.raises(ValueError):
        style.allow_color = 10

    with pytest.raises(ValueError):
        style.allow_color = 3.14159

    style.allow_color = REBECCAPURPLE
    assert_property(style, "allow_color", NAMED_COLOR[REBECCAPURPLE])

    style.allow_color = "#112233"
    assert_property(style, "allow_color", rgb(0x11, 0x22, 0x33))

    with pytest.raises(ValueError):
        style.allow_color = "a"

    with pytest.raises(ValueError):
        style.allow_color = "b"

    # A Travertino NONE is an explicit string value
    with pytest.raises(ValueError):
        style.allow_color = NONE

    # A Python None is invalid
    with pytest.raises(ValueError):
        style.allow_color = None

    # The property can be reset
    del style.allow_color
    assert_property(style, "allow_color", NAMED_COLOR["goldenrod"])

    with pytest.raises(
        ValueError,
        match=(
            r"Invalid value 'invalid' for property allow_color; Valid values are: "
            r"<color>"
        ),
    ):
        style.allow_color = "invalid"


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_values(StyleClass):
    style = StyleClass()
    assert style.values == "a"

    with pytest.raises(ValueError):
        style.values = 10

    with pytest.raises(ValueError):
        style.values = 3.14159

    with pytest.raises(ValueError):
        style.values = REBECCAPURPLE

    with pytest.raises(ValueError):
        style.values = "#112233"

    style.values = NONE
    assert_property(style, "values", NONE)

    style.values = "b"
    assert_property(style, "values", "b")

    # A Python None is invalid
    with pytest.raises(ValueError):
        style.values = None

    # The property can be reset
    del style.values
    assert_property(style, "values", "a")

    with pytest.raises(
        ValueError,
        match=(
            r"Invalid value 'invalid' for property values; Valid values are: a, b, "
            r"none"
        ),
    ):
        style.values = "invalid"


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_multiple_choices(StyleClass):
    style = StyleClass()

    style.multiple_choices = 10
    assert_property(style, "multiple_choices", 10.0)

    style.multiple_choices = 3.14159
    assert_property(style, "multiple_choices", 3.14159)

    style.multiple_choices = REBECCAPURPLE
    assert_property(style, "multiple_choices", NAMED_COLOR[REBECCAPURPLE])

    style.multiple_choices = "#112233"
    assert_property(style, "multiple_choices", rgb(0x11, 0x22, 0x33))

    style.multiple_choices = "a"
    assert_property(style, "multiple_choices", "a")

    style.multiple_choices = NONE
    assert_property(style, "multiple_choices", NONE)

    style.multiple_choices = "b"
    assert_property(style, "multiple_choices", "b")

    # A Python None is invalid
    with pytest.raises(ValueError):
        style.multiple_choices = None

    # The property can be reset
    # There's no initial value, so the property is None
    del style.multiple_choices
    assert style.multiple_choices is None

    # Check the error message
    with pytest.raises(
        ValueError,
        match=(
            r"Invalid value 'invalid' for property multiple_choices; Valid values are: "
            r"a, b, none, <number>, <color>"
        ),
    ):
        style.multiple_choices = "invalid"


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_string_symbol(StyleClass):
    style = StyleClass()

    # Set a symbolic value using the string value of the symbol
    # We can't just use the string directly, though - that would
    # get optimized by the compiler. So we create a string and
    # transform it into the value we want.
    val = "TOP"
    style.string_symbol = val.lower()

    # Both equality and instance checking should work.
    assert_property(style, "string_symbol", TOP)
    assert style.string_symbol is TOP



import pytest

from travertino.constants import (
    BOLD,
    ITALIC,
    NORMAL,
    OBLIQUE,
    SMALL_CAPS,
    SYSTEM_DEFAULT_FONT_SIZE,
)
from travertino.fonts import Font


def assert_font(font, family, size, style, variant, weight):
    assert font.family == family
    assert font.size == size
    assert font.style == style
    assert font.variant == variant
    assert font.weight == weight


@pytest.mark.parametrize(
    "font",
    [
        Font("Comic Sans", "12 pt"),
        Font("Comic Sans", 12),
        Font("Comic Sans", 12, NORMAL, NORMAL, NORMAL),
    ],
)
def test_equality(font):
    assert font == Font("Comic Sans", "12 pt")


@pytest.mark.parametrize(
    "font",
    [
        Font("Comic Sans", 13),
        Font("Comic Sans", 12, ITALIC),
        Font("Times New Roman", 12, NORMAL, NORMAL, NORMAL),
        "a string",
        5,
    ],
)
def test_inqequality(font):
    assert font != Font("Comic Sans", "12 pt")


def test_hash():
    assert hash(Font("Comic Sans", 12)) == hash(Font("Comic Sans", 12))

    assert hash(Font("Comic Sans", 12, weight=BOLD)) != hash(Font("Comic Sans", 12))


@pytest.mark.parametrize(
    "size, kwargs, string",
    [
        (12, {}, "12pt"),
        (12, {"style": ITALIC}, "italic 12pt"),
        (12, {"style": ITALIC, "variant": SMALL_CAPS}, "italic small-caps 12pt"),
        (
            12,
            {"style": ITALIC, "variant": SMALL_CAPS, "weight": BOLD},
            "italic small-caps bold 12pt",
        ),
        (12, {"variant": SMALL_CAPS, "weight": BOLD}, "small-caps bold 12pt"),
        (12, {"weight": BOLD}, "bold 12pt"),
        (12, {"style": ITALIC, "weight": BOLD}, "italic bold 12pt"),
        # Check system default size handling
        (SYSTEM_DEFAULT_FONT_SIZE, {}, "system default size"),
        (SYSTEM_DEFAULT_FONT_SIZE, {"style": ITALIC}, "italic system default size"),
    ],
)
def test_repr(size, kwargs, string):
    assert repr(Font("Comic Sans", size, **kwargs)) == f"<Font: {string} Comic Sans>"


@pytest.mark.parametrize("size", [12, "12", "12pt", "12 pt"])
def test_simple_construction(size):
    assert_font(Font("Comic Sans", size), "Comic Sans", 12, NORMAL, NORMAL, NORMAL)


def test_invalid_construction():
    with pytest.raises(ValueError):
        Font("Comic Sans", "12 quatloos")


@pytest.mark.parametrize(
    "family",
    [
        "Comics Sans",
        "Wingdings",
        "'Comic Sans'",
        '"Comic Sans"',
    ],
)
def test_family(family):
    normalized_family = family.replace("'", "").replace('"', "")
    assert_font(Font(family, 12), normalized_family, 12, NORMAL, NORMAL, NORMAL)


@pytest.mark.parametrize(
    "style, result_style",
    [
        (ITALIC, ITALIC),
        ("italic", ITALIC),
        (OBLIQUE, OBLIQUE),
        ("oblique", OBLIQUE),
        ("something else", NORMAL),
    ],
)
def test_style(style, result_style):
    assert_font(
        Font("Comic Sans", 12, style=style),
        "Comic Sans",
        12,
        result_style,
        NORMAL,
        NORMAL,
    )


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"style": ITALIC},
    ],
)
def test_make_normal_style(kwargs):
    f = Font("Comic Sans", 12, **kwargs)
    assert_font(f.normal_style(), "Comic Sans", 12, NORMAL, NORMAL, NORMAL)


@pytest.mark.parametrize(
    "method, result",
    [
        ("italic", ITALIC),
        ("oblique", OBLIQUE),
    ],
)
def test_make_slanted(method, result):
    f = Font("Comic Sans", 12)
    assert_font(getattr(f, method)(), "Comic Sans", 12, result, NORMAL, NORMAL)


@pytest.mark.parametrize(
    "variant, result",
    [
        (SMALL_CAPS, SMALL_CAPS),
        ("small-caps", SMALL_CAPS),
        ("something else", NORMAL),
    ],
)
def test_variant(variant, result):
    assert_font(
        Font("Comic Sans", 12, variant=variant),
        "Comic Sans",
        12,
        NORMAL,
        result,
        NORMAL,
    )


@pytest.mark.parametrize("kwargs", [{}, {"variant": SMALL_CAPS}])
def test_make_normal_variant(kwargs):
    f = Font("Comic Sans", 12, **kwargs)
    assert_font(f.normal_variant(), "Comic Sans", 12, NORMAL, NORMAL, NORMAL)


def test_make_small_caps():
    f = Font("Comic Sans", 12)
    assert_font(f.small_caps(), "Comic Sans", 12, NORMAL, SMALL_CAPS, NORMAL)


@pytest.mark.parametrize(
    "weight, result",
    [
        (BOLD, BOLD),
        ("bold", BOLD),
        ("something else", NORMAL),
    ],
)
def test_weight(weight, result):
    assert_font(
        Font("Comic Sans", 12, weight=weight),
        "Comic Sans",
        12,
        NORMAL,
        NORMAL,
        result,
    )


@pytest.mark.parametrize("kwargs", [{}, {"weight": BOLD}])
def test_make_normal_weight(kwargs):
    f = Font("Comic Sans", 12, **kwargs)
    assert_font(f.normal_weight(), "Comic Sans", 12, NORMAL, NORMAL, NORMAL)


def test_make_bold():
    f = Font("Comic Sans", 12)
    assert_font(f.bold(), "Comic Sans", 12, NORMAL, NORMAL, BOLD)



from __future__ import annotations

from warnings import filterwarnings, warn

# Make sure deprecation warnings are shown by default
filterwarnings("default", category=DeprecationWarning)


class Condition:
    def __init__(self, **properties):
        """A condition describing the values of one or more properties of a style.

        :param properties: Any number of name: value pairs
        """
        self.properties = properties

    def match(self, style):
        return all(style[name] == value for name, value in self.properties.items())

    def __str__(self):
        return "; ".join(
            f"{name} == {value}" for name, value in self.properties.items()
        )


class aliased_property:
    def __init__(self, source: str | dict, deprecated: bool = False):
        """Create a property that aliases an existing property.

        :param source: If this is a string, it is the name of the property to
            reference. Otherwise, it is a dictionary mapping conditions to the correct
            property name to use. If no condition is met, an AttributeError is raised.
        :deprecated: Is this property name deprecated?
        """
        self.source = source
        self.deprecated = deprecated

    def __set_name__(self, style_class, name):
        self.name = name
        style_class._BASE_ALL_PROPERTIES[style_class].add(name)

    def __get__(self, style, style_class=None):
        if style is None:
            return self

        return style[self.derive_name(style)]

    def __set__(self, style, value):
        if value is self:
            # This happens during autogenerated dataclass __init__ when no value is
            # supplied.
            return

        style[self.derive_name(style)] = value

    def __delete__(self, style):
        del style[self.derive_name(style)]

    def is_set_on(self, style):
        return self.derive_name(style) in style

    def derive_name(self, style):
        name = None

        if isinstance(self.source, str):
            name = self.source
        else:
            for condition, result in self.source.items():
                if condition.match(style):
                    name = result
                    break

        if name is None:
            conditions = " or ".join(f"({condition})" for condition in self.source)
            raise AttributeError(f"'{self.name}' is only supported when {conditions}")

        if self.deprecated:
            cls = type(style).__name__
            warn(
                f"{cls}.{self.name} is deprecated. Use {cls}.{name} instead.",
                DeprecationWarning,
                stacklevel=3,
            )

        return name



from __future__ import annotations

from collections import defaultdict
from collections.abc import Mapping
from contextlib import contextmanager
from warnings import filterwarnings, warn

from .properties.shorthand import directional_property
from .properties.validated import validated_property

# Make sure deprecation warnings are shown by default
filterwarnings("default", category=DeprecationWarning)


class BaseStyle:
    """A base class for style declarations.

    Exposes a dict-like interface. Designed for subclasses to be decorated
    with @dataclass(kw_only=True, repr=False).

    The kw_only parameter was added in Python 3.10; for 3.9, init=False can be used
    instead to still get the keyword-only behavior from the included __init__.

    Most IDEs should see the dataclass decorator and provide autocompletion / type hints
    for parameters to the constructor.
    """

    # Only "real" properties
    _BASE_PROPERTIES = defaultdict(set)
    # Includes aliases and shorthands
    _BASE_ALL_PROPERTIES = defaultdict(set)

    def __init_subclass__(cls):
        # Give the subclass a direct reference to its properties.
        cls._PROPERTIES = cls._BASE_PROPERTIES[cls]
        cls._ALL_PROPERTIES = cls._BASE_ALL_PROPERTIES[cls]

    ########################################################################
    # 03-2025: Backwards compatibility for Toga < 0.5.0 *and* for Python 3.9
    ########################################################################

    # Fallback in case subclass isn't decorated as dataclass (probably from using
    # previous API) or for pre-3.10, before kw_only argument existed.
    def __init__(self, **properties):
        try:
            self.update(**properties)
            self.__post_init__()
        except NameError:
            # It still makes sense for update() to raise a NameError. However, here we
            # simulate the behavior of the dataclass-generated __init__() for
            # consistency.
            for name in properties:
                # This is redoing work, but it should only ever happen when a property
                # name is invalid, and only in outdated Python or Toga, and only once.
                if name not in self._ALL_PROPERTIES:
                    raise TypeError(
                        f"{type(self).__name__}.__init__() got an unexpected keyword "
                        f"argument '{name}'"
                    )
            # The above for loop should never run to completion, so that needs to be
            # excluded from coverage.
            else:  # pragma: no cover
                pass

    ######################################################################
    # End backwards compatibility
    ######################################################################

    @property
    def _applicator(self):
        return getattr(self, "_assigned_applicator", None)

    @_applicator.setter
    def _applicator(self, value):
        self._assigned_applicator = value

        if value is not None:
            try:
                self.apply()
            ######################################################################
            # 10-2024: Backwards compatibility for Toga < 0.5.0
            ######################################################################
            except Exception:
                warn(
                    "Failed to apply style when assigning applicator, or when "
                    "assigning a new style once applicator is present. Node should be "
                    "sufficiently initialized to apply its style before it is assigned "
                    "an applicator. This will be an exception in a future version.\n"
                    "This error probably means you've updated Travertino to 0.5.0 but "
                    "are still using Toga <= 0.4.8; to fix, either update Toga to "
                    ">= 0.5.0, or pin Travertino to 0.3.0.",
                    RuntimeWarning,
                    stacklevel=2,
                )
            ######################################################################
            # End backwards compatibility
            ######################################################################

    def copy(self, applicator=None):
        """Create a duplicate of this style declaration."""
        dup = self.__class__()
        dup.update(**self)

        ######################################################################
        # 10-2024: Backwards compatibility for Toga < 0.5.0
        ######################################################################

        if applicator is not None:
            warn(
                "Providing an applicator to BaseStyle.copy() is deprecated. Set "
                "applicator afterward on the returned copy.\n"
                "This error probably means you've updated Travertino to 0.5.0 but are "
                "still using Toga <= 0.4.8; to fix, either update Toga to >= 0.5.0, or "
                "pin Travertino to 0.3.0.",
                DeprecationWarning,
                stacklevel=2,
            )
            dup._applicator = applicator

        ######################################################################
        # End backwards compatibility
        ######################################################################

        return dup

    ######################################################################
    # Interface that style declarations must define
    ######################################################################

    def _apply(self, names: set):
        raise NotImplementedError(
            "Style must define an _apply method"
        )  # pragma: no cover

    def layout(self, viewport):
        raise NotImplementedError(
            "Style must define a layout method"
        )  # pragma: no cover

    ######################################################################
    # Support for batching calls to apply()
    ######################################################################

    def __post_init__(self):
        # Because batch_apply is a no-op with no applicator, it's fine that these aren't
        # set during the dataclass-generated __init__ — even when directional/composite
        # properties (which call batch_apply) are set.
        self._batched_mode = False
        self._batched_names = set()

    # After deprecation is removed, this should be the signature:
    # def apply(self, name: str | None = None) -> None:
    def apply(self, *names: list[str]) -> None:
        if not self._applicator:
            return

        ######################################################################
        # 03-2025: Backwards compatibility for Toga 0.5.1
        ######################################################################

        if len(names) > 1:
            cls = type(self).__name__
            warn(
                f"Calling {cls}.apply() with multiple arguments is deprecated. Use the "
                f'"with {cls}.batch_apply():" context manager instead.',
                DeprecationWarning,
                stacklevel=2,
            )

        if self._batched_mode:
            self._batched_names.update(names)
        else:
            self._apply({*names} if names else self._PROPERTIES)

        ######################################################################
        # End backwards compatibility
        ######################################################################

        # if self._batched_mode:
        #     self._batched_names.add(name)
        # else:
        #     self._apply({name} if name else self._PROPERTIES)

    @contextmanager
    def batch_apply(self):
        """Aggregate calls to appl() into one single call to _apply().

        No-op if no applicator is present, or if already in batched mode.
        """
        # Short-circuit out if no applicator is set. This avoids trying to access the
        # nonexistent _batched_mode during __init__.
        if batch_entered := self._applicator and not self._batched_mode:
            self._batched_mode = True

        try:
            yield
        finally:
            if batch_entered:
                self._batched_mode = False

                if self._batched_names:
                    self._apply(self._batched_names)
                    self._batched_names.clear()

    ######################################################################
    # Provide a dict-like interface
    ######################################################################

    def update(self, **properties):
        """Set multiple styles on the style definition."""
        # Some aliases may be valid only in the presence of other property values, or
        # depend on other values to determine what to alias to. This update might be
        # setting those prerequisite properties. So we need to defer setting any
        # conditional aliases until last.
        deferred_aliases = {}

        with self.batch_apply():
            for name, value in properties.items():
                name = name.replace("-", "_")
                if name not in self._ALL_PROPERTIES:
                    raise NameError(f"Unknown property '{name}'")

                prop = getattr(type(self), name)
                if isinstance(getattr(prop, "source", None), dict):
                    deferred_aliases[name] = value
                else:
                    self[name] = value

            for name, value in deferred_aliases.items():
                self[name] = value

    def __getitem__(self, name):
        name = name.replace("-", "_")
        if name in self._ALL_PROPERTIES:
            return getattr(self, name)
        raise KeyError(name)

    def __setitem__(self, name, value):
        name = name.replace("-", "_")
        if name in self._ALL_PROPERTIES:
            setattr(self, name, value)
        else:
            raise KeyError(name)

    def __delitem__(self, name):
        name = name.replace("-", "_")
        if name in self._ALL_PROPERTIES:
            delattr(self, name)
        else:
            raise KeyError(name)

    def keys(self):
        return {*self}

    def items(self):
        return [(name, self[name]) for name in self]

    def __len__(self):
        return sum(1 for _ in self)

    def __contains__(self, name):
        name = name.replace("-", "_")
        return name in self._ALL_PROPERTIES and (
            getattr(self.__class__, name).is_set_on(self)
        )

    def __iter__(self):
        yield from (name for name in self._PROPERTIES if name in self)

    def __or__(self, other):
        if not (type(self) is type(other) or isinstance(other, Mapping)):
            return NotImplemented

        result = self.copy()
        result.update(**other)
        return result

    def __ior__(self, other):
        if not (type(self) is type(other) or isinstance(other, Mapping)):
            return NotImplemented

        self.update(**other)
        return self

    ######################################################################
    # Get the rendered form of the style declaration
    ######################################################################

    def __str__(self):
        return "; ".join(
            f"{name.replace('_', '-')}: {value}" for name, value in sorted(self.items())
        )

    def __repr__(self):
        properties = ", ".join(
            f"{name}={repr(value)}" for name, value in sorted(self.items())
        )
        return f"{type(self).__name__}({properties})"

    ######################################################################
    # Backwards compatibility
    ######################################################################

    def reapply(self):
        warn(
            "BaseStyle.reapply() is deprecated; call .apply with no arguments "
            "instead.\n"
            "This error probably means you've updated Travertino to 0.5.0 but are "
            "still using Toga <= 0.4.8; to fix, either update Toga to >= 0.5.0, or pin "
            "Travertino to 0.3.0.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.apply()

    @classmethod
    def validated_property(cls, name, choices, initial=None):
        warn(
            "Defining style properties with class methods is deprecated; use class "
            "attributes instead.\n"
            "This error probably means you've updated Travertino to 0.5.0 but are "
            "still using Toga <= 0.4.8; to fix, either update Toga to >= 0.5.0, or pin "
            "Travertino to 0.3.0.",
            DeprecationWarning,
            stacklevel=2,
        )
        prop = validated_property(
            *choices.constants,
            string=choices.string,
            integer=choices.integer,
            number=choices.number,
            color=choices.color,
            initial=initial,
        )
        setattr(cls, name, prop)
        prop.__set_name__(cls, name)

    @classmethod
    def directional_property(cls, name):
        warn(
            "Defining style properties with class methods is deprecated; use class "
            "attributes instead.\n"
            "This error probably means you've updated Travertino to 0.5.0 but are "
            "still using Toga <= 0.4.8; to fix, either update Toga to >= 0.5.0, or pin "
            "Travertino to 0.3.0.",
            DeprecationWarning,
            stacklevel=2,
        )
        name_format = name % "{}"
        name = name_format.format("")
        prop = directional_property(name_format)
        setattr(cls, name, prop)
        prop.__set_name__(cls, name)



from .constants import (
    BOLD,
    FONT_STYLES,
    FONT_VARIANTS,
    FONT_WEIGHTS,
    ITALIC,
    NORMAL,
    OBLIQUE,
    SMALL_CAPS,
    SYSTEM_DEFAULT_FONT_SIZE,
)


class Font:
    def __init__(self, family, size, style=NORMAL, variant=NORMAL, weight=NORMAL):
        if (family[0] == "'" and family[-1] == "'") or (
            family[0] == '"' and family[-1] == '"'
        ):
            self.family = family[1:-1]
        else:
            self.family = family

        try:
            self.size = int(size)
        except ValueError:
            try:
                if size.strip().endswith("pt"):
                    self.size = int(size[:-2])
                else:
                    raise ValueError(f"Invalid font size {size!r}")
            except Exception:
                raise ValueError(f"Invalid font size {size!r}")
        self.style = style if style in FONT_STYLES else NORMAL
        self.variant = variant if variant in FONT_VARIANTS else NORMAL
        self.weight = weight if weight in FONT_WEIGHTS else NORMAL

    def __hash__(self):
        return hash(
            ("FONT", self.family, self.size, self.style, self.variant, self.weight)
        )

    def __repr__(self):
        return "<Font: {}{}{}{} {}>".format(
            "" if self.style is NORMAL else (self.style + " "),
            "" if self.variant is NORMAL else (self.variant + " "),
            "" if self.weight is NORMAL else (self.weight + " "),
            (
                "system default size"
                if self.size == SYSTEM_DEFAULT_FONT_SIZE
                else f"{self.size}pt"
            ),
            self.family,
        )

    def __eq__(self, other):
        try:
            return (
                self.family == other.family
                and self.size == other.size
                and self.style == other.style
                and self.variant == other.variant
                and self.weight == other.weight
            )
        except AttributeError:
            return False

    def normal_style(self):
        "Generate a normal style version of this font"
        return Font(
            self.family,
            self.size,
            style=NORMAL,
            variant=self.variant,
            weight=self.weight,
        )

    def italic(self):
        "Generate an italic version of this font"
        return Font(
            self.family,
            self.size,
            style=ITALIC,
            variant=self.variant,
            weight=self.weight,
        )

    def oblique(self):
        "Generate an oblique version of this font"
        return Font(
            self.family,
            self.size,
            style=OBLIQUE,
            variant=self.variant,
            weight=self.weight,
        )

    def normal_variant(self):
        "Generate a normal variant of this font"
        return Font(
            self.family, self.size, style=self.style, variant=NORMAL, weight=self.weight
        )

    def small_caps(self):
        "Generate a small-caps variant of this font"
        return Font(
            self.family,
            self.size,
            style=self.style,
            variant=SMALL_CAPS,
            weight=self.weight,
        )

    def normal_weight(self):
        "Generate a normal weight version of this font"
        return Font(
            self.family,
            self.size,
            style=self.style,
            variant=self.variant,
            weight=NORMAL,
        )

    def bold(self):
        "Generate a bold version of this font"
        return Font(
            self.family, self.size, style=self.style, variant=self.variant, weight=BOLD
        )



class at_least:
    """An annotation to wrap around a value to describe that it is a minimum bound"""

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f"at least {self.value}"

    def __eq__(self, other):
        try:
            return self.value == other.value
        except AttributeError:
            return False


class BaseIntrinsicSize:
    """Representation of the intrinsic size of an object.

    width: The width of the node.
    height: The height of the node.
    """

    def __init__(self, width=None, height=None):
        self.width = width
        self.height = height

    def __repr__(self):
        return f"({self.width}, {self.height})"



from ..constants import BOTTOM, LEFT, RIGHT, TOP


class directional_property:
    ASSIGNMENT_SCHEMES = {
        #   T  R  B  L
        1: [0, 0, 0, 0],
        2: [0, 1, 0, 1],
        3: [0, 1, 2, 1],
        4: [0, 1, 2, 3],
    }

    def __init__(self, name_format):
        """Define a property that proxies for top/right/bottom/left alternatives.

        :param name_format: The format from which to generate subproperties. "{}" will
            be replaced with "_top", etc.
        """
        self.property_names = [
            name_format.format(f"_{direction}")
            for direction in [TOP, RIGHT, BOTTOM, LEFT]
        ]

    def __set_name__(self, style_class, name):
        self.name = name
        style_class._BASE_ALL_PROPERTIES[style_class].add(name)

    def __get__(self, style, style_class=None):
        if style is None:
            return self

        return tuple(style[name] for name in self.property_names)

    def __set__(self, style, value):
        if value is self:
            # This happens during autogenerated dataclass __init__ when no value is
            # supplied.
            return

        if not isinstance(value, tuple):
            value = (value,)

        if order := self.ASSIGNMENT_SCHEMES.get(len(value)):
            with style.batch_apply():
                for name, index in zip(self.property_names, order):
                    style[name] = value[index]
        else:
            raise ValueError(
                f"Invalid value for '{self.name}'; value must be a number, or a 1-4 "
                f"tuple."
            )

    def __delete__(self, style):
        with style.batch_apply():
            for name in self.property_names:
                del style[name]

    def is_set_on(self, style):
        return any(name in style for name in self.property_names)
